using Toybox.Application as App;
using Toybox.Application.Storage;
using Toybox.Graphics as Gfx;
using Toybox.Math;
using Toybox.System as Sys;
using Toybox.Time;
using Toybox.Time.Gregorian as Cal;
using Toybox.Weather;

import Toybox.Lang;

module C
{

//
//  Constants
//
public var WHITE    = 0xffffff;     // Gfx.COLOR_WHITE;
public var BLACK    = 0x000000;     // Gfx.COLOR_BLACK;
public var RED      = 0xff0000;     // Gfx.COLOR_RED;
public var GREEN    = 0x00ff00;     // Gfx.COLOR_GREEN;
public var BLUE     = 0x0000ff;     // Gfx.COLOR_BLUE;
public var CYAN     = 0x00ffff;
public var MAGENTA  = 0xff00ff;
public var YELLOW   = 0xffff00;     // Gfx.COLOR_YELLOW;
public var ORANGE   = 0xff8800;     // Gfx.COLOR_ORANGE;
public var TRANSPARENT = Gfx.COLOR_TRANSPARENT;

public var CACHE_MAX = 8;       // maximum size of the TZ cache

//
//  Settings variables
//
public var app_ver;
public var ampm;
public var secm;
public var sec_idx;
public var sec_col;
public var ccw;
public var backing;
public var quad_ul;
public var quad_ur;
public var quad_lr;
public var quad_ll;
public var temp;
public var tz1 = 0;
public var tz1_dst;
public var tz2 = 0;
public var tz2_dst;

public var upd_back = false;
public var sleeping = false;

//
//  Dimensions of the watch
//
public var radius;
public var center_x;
public var center_y;
public var dot_sz;

//
// Timezone info
//
public var tz1_info as Array<String or Number> or Null;
public var tz2_info as Array<String or Number> or Null;

//
//  Debug info
//
(:debug)   var debug = true;
(:release) var debug = false;
public var dbg_ok = false;

//
//  Debugging moon phase
//
var dbg_moon = false;
var dbg_time;
var dbg_dsecs;
var dbg_msecs;
var dbg_secs;

//
//  Menu entries
//
public enum {
    MENU_IGNORE,
    MENU_TZ1,
    MENU_TZ2,
    MENU_CONTROL,
    MENU_SECOND,
    MENU_AMPM,
    MENU_CCW,
    MENU_BACKING,
    MENU_CACHE,
    MENU_SHAND,
    MENU_SCOLOR,
    MENU_TZ_CACHE,
    MENU_TZ_USA,
    MENU_TZ_A,
    MENU_TZ_B,
    MENU_TZ_C,
    MENU_TZ_D,
    MENU_TZ_E,
    MENU_TZ_F,
    MENU_TZ_G,
    MENU_TZ_H,
    MENU_TZ_I,
    MENU_TZ_J,
    MENU_TZ_K,
    MENU_TZ_L,
    MENU_TZ_M,
    MENU_TZ_N,
    MENU_TZ_O,
    MENU_TZ_P,
    MENU_TZ_Q,
    MENU_TZ_R,
    MENU_TZ_S,
    MENU_TZ_T,
    MENU_TZ_U,
    MENU_TZ_V,
    MENU_TZ_W,
    MENU_TZ_X,
    MENU_TZ_Y,
    MENU_TZ_Z,
    MENU_ZONES      // this MUST be last with largest value
}

//
//  These enums MUST match the constants for the 4
//  settings quad_ul, quad_ur, quad_ll, quad_lr
//  in settings.xml
//
public enum {
    DATA_NONE = 0,
    DATA_BATTERY,
    DATA_HEART,
    DATA_MOON,
    DATA_STEPS,
    DATA_CALORIES,
    DATA_DIST,
    DATA_ASCENT,
    DATA_ACTIVE,
    DATA_RESP,
    DATA_TEMP
}

public enum {
    GRAVITY_UL,
    GRAVITY_UR,
    GRAVITY_LR,
    GRAVITY_LL
}

//
//  These enums MUST match the constants for
//  sec_col in settings.xml
//
public enum {
    S_RED = 0,
    S_GREEN,
    S_BLUE,
    S_CYAN,
    S_MAGENTA,
    S_YELLOW,
    S_WHITE,
    S_CUSTOM
}

public function logger(s)
{

    if (debug) {
        Sys.println("DDD:" + s);
    }
}

public function pr_date(t, d)
{

    if (d instanceof Time.Moment) {
        d = Cal.info(d, Time.FORMAT_SHORT); }
    C.logger(t + ":" + d.month + "/" + d.day + "/" + d.year + " " +
                       d.hour +  ":" + d.min.format("%02d") + ":" + d.sec.format("%02d") +
                       utc_off(Sys.getClockTime().timeZoneOffset));
}

public function pr_memory(pre, suf)
{
    var total, used, free;
    var scale;

    if (suf == null) {
        suf = 'u'; }

    switch (suf) {

    case 'k':   case 'K':
        scale = 1000;
        break;

    case 'm':   case 'M':
        scale = 1000*1000;
        break;

    case 'g':   case 'G':
        scale = 1000*1000*1000;
        break;

    default:
        scale = 1;
        suf = "";
        break;

    }

    var stats = Sys.getSystemStats();
    total = stats.totalMemory / scale;
    used  = stats.usedMemory / scale;
    free  = stats.freeMemory / scale;
    logger(pre + ":" + total + suf + "=" + used + suf + "+" + free + suf);
}

public function pr_mem(pre)
{

    pr_memory(pre, 'k');
}

public function pr_xy(xy as Array<Array>)
{
    var v;

    var len = xy.size();
    for (var i = 0; i < len; i += 1) {
        v = xy[i];
        logger("xy[i] = " + v[0] + ", " + v[1]);
    }
}

function to_cart(r, a) as Array<Number>
{

    var x = r * Math.cos(a);
    var y = r * Math.sin(a);
    return [ x, y ];
}

function to_polar(x, y) as Array<Number>
{

    var r = Math.sqrt((x*x) + (y*y));
    var a = Math.atan2(x, y);
    return [ r, a ];
}

function translate(vert as Array<Array<Number>>, x, y, a) as Array<Array<Number>>
{
    var vs, p, c;

    var l = vert.size();
    var v = new [l];
    for (var i = 0; i < l; i += 1) {
        vs = vert[i];
        p = C.to_polar(vs[0], vs[1]);
        c = C.to_cart(p[0], p[1] + a);
        c[0] += C.center_x;
        c[1] += C.center_y;
        v[i] = c;
    }
    return v;
}

function rotate(len, tick) as Array<Number>
{

    var v = new [1];
    v[0] = [ 0, len ];
    var m = tick;
    var a = (((Math.PI * 2) * m) / 60) + (Math.PI/2);
    return C.translate(v, C.center_x, C.center_y, a)[0];
}

function line(dc, p1 as Array<Number>, p2 as Array<Number>)
{

    dc.drawLine(p1[0], p1[1], p2[0], p2[1]);
}

function alpha(c, a)
{

    return c | (a << 24);
}

function utc(now, off)
{

    now = now.subtract(new Time.Duration(Sys.getClockTime().timeZoneOffset));
    now = now.add(new Time.Duration(off));
    return now;
}

function utc_off(sec)
{
    var pre;

    if (sec < 0) {
        sec = -sec;
        pre = " -";
    } else {
        pre = " +"; }

    var min = (sec / 60).toNumber();
    var h = (min / 60).toNumber();
    return pre + h + ":" + (min - (h*60)).format("%02d");
}

function get_prop(name, def as String or Number or Boolean) as String or Number or Boolean or Null
{

    var v = App.Properties.getValue(name);
    return (v == null) ? def : v;
}

function set_prop(name, value as String or Number or Boolean or Null)
{

    App.Properties.setValue(name, value);
}

//
//  Properties:
//
//    ampm - what AM/PM indication to disply
//    ccw - hands run counter clockwise
//    tz1 - main timezone
//    tz2 - secondary  timezone
//    cache - timezone cache
//
function settings()
{

    app_ver = get_prop("version", "-.-");
    dbg_ok = get_prop("dbg_enable", false);
    ampm = get_prop("ampm", 0);
    secm = get_prop("secm", 0);
    sec_idx = get_prop("sec_col", S_RED);
    switch (sec_idx) {

    default:
    case S_RED:
        sec_col = RED;
        break;

    case S_GREEN:
        sec_col = GREEN;
        break;

    case S_BLUE:
        sec_col = BLUE;
        break;

    case S_CYAN:
        sec_col = CYAN;
        break;

    case S_MAGENTA:
        sec_col = MAGENTA;
        break;

    case S_YELLOW:
        sec_col = YELLOW;
        break;

    case S_WHITE:
        sec_col = WHITE;
        break;

    case S_CUSTOM:
        sec_col = s2i(get_prop("col_cust", 0), 16);
        break;
    }

    ccw = get_prop("ccw", false) ? -1 : 1;
    var b = get_prop("backing", false);
    if (b != backing) {
        upd_back = true; }
    backing = b;
    quad_ul = get_prop("quad_ul", DATA_NONE);
    quad_ur = get_prop("quad_ur", DATA_NONE);
    quad_lr = get_prop("quad_lr", DATA_NONE);
    quad_ll = get_prop("quad_ll", DATA_NONE);
    temp = get_prop("temp", -999);
//quad_ur = C.DATA_DIST;
//quad_ll = C.DATA_RESP;
    tz1 = get_prop("tz1", 0);
    tz1_dst = get_prop("tz1_dst", false);
    tz2 = get_prop("tz2", 0);
    tz2_dst = get_prop("tz2_dst", false);
pr_date("settings", get_now());
logger("         am/pm - " + ampm);
logger("         secm  - " + secm);
logger("         sec_c - " + sec_col.format("%06x"));
logger("         ccw   - " + ccw);
logger("         back  - " + backing);
logger("         tz1   - " + tz1);
logger("         tz2   - " + tz2);
logger("         dbg   - " + dbg_ok);

    tz1_info = C.tz_get(tz1, tz1_dst);
    tz2_info = C.tz_get(tz2, tz2_dst);

    var cache = new TzCache();

    cache.push(tz1_info[4], tz1);
    cache.push(tz2_info[4], tz2);
    cache = null;

    pr_mem("         mem");
}

function do_dbg(partial, bitm, sec)
{

    if (dbg_ok) {
        var str = partial + "," + bitm + "," + sec;
        set_prop("dbg_line", str);
    }
}

//
//  Note - we don't have to set the global variables for
//  the properties, this routine will always be folowed
//  by a call to settings() which will set the global
//  variables at that time.
function reconfig(p_tz1, p_tz2, p_ampm, p_ccw, p_secm, p_sec_idx, p_backing, p_tz_cache)
{

    //C.logger("reconfig:" + p_tz_cache + ", " + p_tz1 + ". " + p_tz2);

    if (p_tz_cache) {
        var cache = new TzCache();
        cache.reset();
        tz1 = 0;
        tz2 = 0;
    }

    set_prop("tz1", p_tz1);
    set_prop("tz2", p_tz2);
    set_prop("ampm", p_ampm);
    set_prop("ccw", p_ccw == -1);
    set_prop("secm", p_secm);
    set_prop("sec_col", p_sec_idx);
    set_prop("backing", p_backing);

    Storage.setValue("TZ1old", p_tz1);
    Storage.setValue("TZ2old", p_tz2);

    settings();
}

function time_adj(now, tz_info as Array<Number>)
{

//var info = Cal.info(Time.now(), Time.FORMAT_SHORT);
//C.logger("time_adj:" + info.hour + ":" + info.min + " " + info.day + ", " + info.day_of_week);
//C.logger("time_adj:" + tz_info[0] + ", " + tz_info[1] + ", " + tz_info[2] + ", " + tz_info[3]);
    if (tz_info[0] != 0) {
        now = utc(now, tz_info[0]*60); }

    var time = new MyTime(now, tz_info);
    return time;
}

function get_temp(now) as Array<String or Number>
{

    var w = Weather.getCurrentConditions();
    if (w == null) {
        return [ "--", Rez.Drawables.temp, WHITE ]; }
    var t = w.temperature;
    if (t == null) {
        return [ "-", Rez.Drawables.temp, WHITE ]; }

    if (temp == 1) {
        t = ((t * 9.0)/5.0) + 32.0; }

    var delta = now.value() - w.observationTime.value();
    var icon = Rez.Drawables.temp_red;
    var color = RED;
    if (delta < ((30*60) + (5*60))) {
        icon = Rez.Drawables.temp_green;
        color = GREEN;
    } else if (delta < ((1*60*60) + (5*60))) {
        icon = Rez.Drawables.temp_yellow;
        color = YELLOW;
    } else if (delta < ((2*60*60) + (5*60))) {
        icon = Rez.Drawables.temp_orange;
        color = ORANGE;
    }

    return [ t.format("%.0f"), icon, color ];
}

function get_now()
{

    var now = Time.now();
    if (!dbg_moon) {
        return now; }

    var DAY_SECS = 60 * 60 * 24;
    if (dbg_dsecs == null) {
        dbg_dsecs = (now.value()/DAY_SECS).toNumber();
        dbg_msecs = 0;
        dbg_secs = 0;
    }
    if (now.value() != dbg_time) {
        dbg_time = now.value();
        dbg_secs += 1;
        if (dbg_secs >= 60) {
            dbg_secs = 0; }
    }
    if (dbg_secs == 4) {
        dbg_dsecs += 1;
        dbg_msecs += 1;
        dbg_secs = 59;
    }
    var nt = (dbg_dsecs * DAY_SECS) + (dbg_msecs * 60) + dbg_secs;
//var tinfo = Cal.info(new Time.Moment(nt), Time.FORMAT_SHORT);
//C.logger("get_now:" + tinfo.min + ":" + tinfo.sec + " => " + nt);
    return new Time.Moment(nt);
}

function set_fill(dc, color)
{

    if (dc has :setFill) {
        dc.setFill(color);
    } else {
        dc.setColor(color, TRANSPARENT);
    }
}

function s2i(str, base)
{

    if (str == null) {
        return 0; }
    return str.toLongWithBase(base).toNumber();
}

function tz_data(idx)
{

         if (idx >= TzOffsets.off[25]) { return new TzZ().TzData[idx - TzOffsets.off[25]]; }
    else if (idx >= TzOffsets.off[24]) { return new TzY().TzData[idx - TzOffsets.off[24]]; }
    else if (idx >= TzOffsets.off[23]) { return new TzX().TzData[idx - TzOffsets.off[23]]; }
    else if (idx >= TzOffsets.off[22]) { return new TzW().TzData[idx - TzOffsets.off[22]]; }
    else if (idx >= TzOffsets.off[21]) { return new TzV().TzData[idx - TzOffsets.off[21]]; }
    else if (idx >= TzOffsets.off[20]) { return new TzU().TzData[idx - TzOffsets.off[20]]; }
    else if (idx >= TzOffsets.off[19]) { return new TzT().TzData[idx - TzOffsets.off[19]]; }
    else if (idx >= TzOffsets.off[18]) { return new TzS().TzData[idx - TzOffsets.off[18]]; }
    else if (idx >= TzOffsets.off[17]) { return new TzR().TzData[idx - TzOffsets.off[17]]; }
    else if (idx >= TzOffsets.off[16]) { return new TzQ().TzData[idx - TzOffsets.off[16]]; }
    else if (idx >= TzOffsets.off[15]) { return new TzP().TzData[idx - TzOffsets.off[15]]; }
    else if (idx >= TzOffsets.off[14]) { return new TzO().TzData[idx - TzOffsets.off[14]]; }
    else if (idx >= TzOffsets.off[13]) { return new TzN().TzData[idx - TzOffsets.off[13]]; }
    else if (idx >= TzOffsets.off[12]) { return new TzM().TzData[idx - TzOffsets.off[12]]; }
    else if (idx >= TzOffsets.off[11]) { return new TzL().TzData[idx - TzOffsets.off[11]]; }
    else if (idx >= TzOffsets.off[10]) { return new TzK().TzData[idx - TzOffsets.off[10]]; }
    else if (idx >= TzOffsets.off[9])  { return new TzJ().TzData[idx - TzOffsets.off[9]]; }
    else if (idx >= TzOffsets.off[8])  { return new TzI().TzData[idx - TzOffsets.off[8]]; }
    else if (idx >= TzOffsets.off[7])  { return new TzH().TzData[idx - TzOffsets.off[7]]; }
    else if (idx >= TzOffsets.off[6])  { return new TzG().TzData[idx - TzOffsets.off[6]]; }
    else if (idx >= TzOffsets.off[5])  { return new TzF().TzData[idx - TzOffsets.off[5]]; }
    else if (idx >= TzOffsets.off[4])  { return new TzE().TzData[idx - TzOffsets.off[4]]; }
    else if (idx >= TzOffsets.off[3])  { return new TzD().TzData[idx - TzOffsets.off[3]]; }
    else if (idx >= TzOffsets.off[2])  { return new TzC().TzData[idx - TzOffsets.off[2]]; }
    else if (idx >= TzOffsets.off[1])  { return new TzB().TzData[idx - TzOffsets.off[1]]; }
    else if (idx >= TzOffsets.off[0])  { return new TzA().TzData[idx - TzOffsets.off[0]]; }
    return new Tz0().TzData[idx];
}

public function tz_get(idx, dst) as Array<String or Number>
{

    var data = tz_data(idx);
    var i = data.find(";");
    var name = data.substring(0, i);

    data = data.substring(i + 1, data.length());
    i = data.find(";");
    var off = data.substring(0, i).toNumber();

    data = data.substring(i + 1, data.length());
    i = data.find(";");
    var st_ab = data.substring(0, i);

    data = data.substring(i + 1, data.length());
    i = data.find(";");
    var dst_ab = data.substring(0, i);

    data = data.substring(i + 1, data.length());
    var auto = data.substring(0, data.length()).toNumber();
    if (dst > 0) {
        auto = (dst == 1) ? -2 : -3; }

    return [ off, st_ab, dst_ab, auto, name ];
}

}
